#include "main.h"
#include "APF.h"

#include <vector>

std::vector<GridPoint> startpoints, endpoints;
std::vector<unsigned char> returnpath;
struct ReturnGridPoint {
	unsigned long gridid;
	unsigned long x, y;
	ReturnGridPoint() {
		gridid = 0;
		x = 0;
		y = 0;
	}
} returnstartpoint, returnendpoint;
unsigned long returncost;

gmexport double apf_grid_create(double width, double height) {
	Grid *grid = new Grid(gm_cast<unsigned long>(width), gm_cast<unsigned long>(height));
	return grid->GetID();
}

gmexport double apf_grid_destroy(double id) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	delete grid;
	return 1;
}

gmexport double apf_grid_clear(double id) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	grid->Clear();
	return 1;
}

/*gmexport double apf_grid_get_bottom_border(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	if(xx>=0 && xx<(long)(grid->GetWidth()) && yy>=0 && yy<(long)(grid->GetHeight())) {
	
}*/

gmexport double apf_grid_get_cell_cost(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>=0 && xx<(long)(grid->GetWidth()) && yy>=0 && yy<(long)(grid->GetHeight())) {
		return grid->GetCell(xx, yy).cost;
	}
	return 0;
}

gmexport double apf_grid_set_cell_cost(double id, double x, double y, double cost) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>=0 && xx<(long)(grid->GetWidth()) && yy>=0 && yy<(long)(grid->GetHeight())) {
		grid->GetCell(xx, yy).cost = gm_cast<unsigned long>(cost);
	}
	return 1;
}

gmexport double apf_grid_get_bottom_border(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>=0 && xx<(long)(grid->GetWidth()) && yy>=0 && yy<(long)(grid->GetHeight())) {
		return grid->GetCell(xx, yy).bottomborder;
	}
	return 0;
}

gmexport double apf_grid_get_right_border(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>=0 && xx<(long)(grid->GetWidth()) && yy>=0 && yy<(long)(grid->GetHeight())) {
		return grid->GetCell(xx, yy).rightborder;
	}
	return 0;
}

gmexport double apf_grid_get_bottom_right_corner(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>=0 && xx<(long)(grid->GetWidth()) && yy>=0 && yy<(long)(grid->GetHeight())) {
		return grid->GetCell(xx, yy).bottomrightcorner;
	}
	return 0;
}

gmexport double apf_grid_add_cell(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>0 && xx<(long)(grid->GetWidth())) {
		if(yy>0 && yy<(long)(grid->GetHeight())) {
			++(grid->GetCell(xx-1, yy-1).bottomrightcorner);
		}
		if(yy>=0 && yy<(long)(grid->GetHeight())) {
			++(grid->GetCell(xx-1, yy).rightborder);
			++(grid->GetCell(xx-1, yy).bottomrightcorner);
		}
	}
	if(xx>=0 && xx<(long)(grid->GetWidth())) {
		if(yy>0 && yy<(long)(grid->GetHeight())) {
			++(grid->GetCell(xx, yy-1).bottomborder);
			++(grid->GetCell(xx, yy-1).bottomrightcorner);
		}
		if(yy>=0 && yy<(long)(grid->GetHeight())) {
			++(grid->GetCell(xx, yy).bottomborder);
			++(grid->GetCell(xx, yy).rightborder);
			++(grid->GetCell(xx, yy).bottomrightcorner);
		}
	}
	return 1;
}

gmexport double apf_grid_remove_cell(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>0 && xx<(long)(grid->GetWidth())) {
		if(yy>0 && yy<(long)(grid->GetHeight())) {
			--(grid->GetCell(xx-1, yy-1).bottomrightcorner);
		}
		if(yy>=0 && yy<(long)(grid->GetHeight())) {
			--(grid->GetCell(xx-1, yy).rightborder);
			--(grid->GetCell(xx-1, yy).bottomrightcorner);
		}
	}
	if(xx>=0 && xx<(long)(grid->GetWidth())) {
		if(yy>0 && yy<(long)(grid->GetHeight())) {
			--(grid->GetCell(xx, yy-1).bottomborder);
			--(grid->GetCell(xx, yy-1).bottomrightcorner);
		}
		if(yy>=0 && yy<(long)(grid->GetHeight())) {
			--(grid->GetCell(xx, yy).bottomborder);
			--(grid->GetCell(xx, yy).rightborder);
			--(grid->GetCell(xx, yy).bottomrightcorner);
		}
	}
	return 1;
}

gmexport double apf_grid_add_line(double id, double x1, double y1, double x2, double y2) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx1 = gm_cast<long>(x1);
	long yy1 = gm_cast<long>(y1);
	long xx2 = gm_cast<long>(x2);
	long yy2 = gm_cast<long>(y2);
	if(yy1>0 && yy1<(long)(grid->GetHeight())) {
		unsigned long xmin = (xx1>0)? xx1 : 0;
		unsigned long xmax = (xx2<(long)(grid->GetWidth()))? ((xx2>0)? xx2 : 0) : grid->GetWidth();
		if(xmin>0 && xmin<grid->GetWidth()) {
			++(grid->GetCell(xmin-1, yy1-1).bottomrightcorner);
		}
		for(unsigned long xx = xmin; xx<xmax; ++xx) {
			++(grid->GetCell(xx, yy1-1).bottomborder);
			++(grid->GetCell(xx, yy1-1).bottomrightcorner);
		}
	}
	if(xx2>0 && xx2<(long)(grid->GetWidth())) {
		unsigned long ymin = (yy1>0)? yy1 : 0;
		unsigned long ymax = (yy2<(long)(grid->GetHeight()))? ((yy2>0)? yy2 : 0) : grid->GetHeight();
		if(ymin>0 && ymin<grid->GetHeight()) {
			++(grid->GetCell(xx2-1, ymin-1).bottomrightcorner);
		}
		for(unsigned long yy = ymin; yy<ymax; ++yy) {
			++(grid->GetCell(xx2-1, yy).rightborder);
			++(grid->GetCell(xx2-1, yy).bottomrightcorner);
		}
	}
	return 1;
}

gmexport double apf_grid_remove_line(double id, double x1, double y1, double x2, double y2) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx1 = gm_cast<long>(x1);
	long yy1 = gm_cast<long>(y1);
	long xx2 = gm_cast<long>(x2);
	long yy2 = gm_cast<long>(y2);
	if(yy1>0 && yy1<(long)(grid->GetHeight())) {
		unsigned long xmin = (xx1>0)? xx1 : 0;
		unsigned long xmax = (xx2<(long)(grid->GetWidth()))? ((xx2>0)? xx2 : 0) : grid->GetWidth();
		if(xmin>0 && xmin<grid->GetWidth()) {
			--(grid->GetCell(xmin-1, yy1-1).bottomrightcorner);
		}
		for(unsigned long xx = xmin; xx<xmax; ++xx) {
			--(grid->GetCell(xx, yy1-1).bottomborder);
			--(grid->GetCell(xx, yy1-1).bottomrightcorner);
		}
	}
	if(xx2>0 && xx2<(long)(grid->GetWidth())) {
		unsigned long ymin = (yy1>0)? yy1 : 0;
		unsigned long ymax = (yy2<(long)(grid->GetHeight()))? ((yy2>0)? yy2 : 0) : grid->GetHeight();
		if(ymin>0 && ymin<grid->GetHeight()) {
			--(grid->GetCell(xx2-1, ymin-1).bottomrightcorner);
		}
		for(unsigned long yy = ymin; yy<ymax; ++yy) {
			--(grid->GetCell(xx2-1, yy).rightborder);
			--(grid->GetCell(xx2-1, yy).bottomrightcorner);
		}
	}
	return 1;
}

gmexport double apf_add_startpoint(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>=0 && xx<(long)(grid->GetWidth()) && yy>=0 && yy<(long)(grid->GetHeight())) {
		startpoints.push_back(GridPoint(grid, xx, yy));
	}
	return 1;
}

gmexport double apf_add_endpoint(double id, double x, double y) {
	Grid *grid = Grid::Find(gm_cast<unsigned long>(id));
	if(grid==NULL) return 0;
	long xx = gm_cast<long>(x);
	long yy = gm_cast<long>(y);
	if(xx>=0 && xx<(long)(grid->GetWidth()) && yy>=0 && yy<(long)(grid->GetHeight())) {
		endpoints.push_back(GridPoint(grid, xx, yy));
	}
	return 1;
}

gmexport char* apf_find_path(double cost_normal, double cost_diagonal) {
	GridPoint startpoint, endpoint;
	if(!Grid::FindPath(startpoints, endpoints, &returnpath, &startpoint, &endpoint, &returncost, gm_cast<unsigned long>(cost_normal), gm_cast<unsigned long>(cost_diagonal))) {
		startpoints.clear();
		endpoints.clear();
		return "";
	}
	startpoints.clear();
	endpoints.clear();
	returnstartpoint.gridid = startpoint.grid->GetID();
	returnstartpoint.x = startpoint.x;
	returnstartpoint.y = startpoint.y;
	returnendpoint.gridid = endpoint.grid->GetID();
	returnendpoint.x = endpoint.x;
	returnendpoint.y = endpoint.y;
	returnpath.push_back('\0');
	return (char*)(&(returnpath.front()));
}

gmexport double apf_get_startpoint_grid() {
	return returnstartpoint.gridid;
}
gmexport double apf_get_startpoint_x() {
	return returnstartpoint.x;
}
gmexport double apf_get_startpoint_y() {
	return returnstartpoint.y;
}
gmexport double apf_get_endpoint_grid() {
	return returnstartpoint.gridid;
}
gmexport double apf_get_endpoint_x() {
	return returnendpoint.x;
}
gmexport double apf_get_endpoint_y() {
	return returnendpoint.y;
}
gmexport double apf_get_cost() {
	return returncost;
}

gmexport double apf_passage_create(double grid1_id, double x1, double y1, double cost1, double grid2_id, double x2, double y2, double cost2, double dir) {
	Grid *grid1 = Grid::Find(gm_cast<unsigned long>(grid1_id));
	if(grid1==NULL) return 0;
	Grid *grid2 = Grid::Find(gm_cast<unsigned long>(grid2_id));
	if(grid2==NULL) return 0;
	long xx1 = gm_cast<long>(x1);
	long yy1 = gm_cast<long>(y1);
	long xx2 = gm_cast<long>(x2);
	long yy2 = gm_cast<long>(y2);
	long d = gm_cast<long>(dir);
	if(d!=2 && d!=4 && d!=6 && d!=8) return 0;
	Passage *passage = new Passage(GridPoint(grid1, xx1, yy1), GridPoint(grid2, xx2, yy2), gm_cast<unsigned long>(cost1), gm_cast<unsigned long>(cost2), d);
	return passage->GetID();
}

gmexport double apf_passage_destroy(double id) {
	Passage *passage = Passage::Find(gm_cast<unsigned long>(id));
	if(passage==NULL) return 0;
	delete passage;
	return 1;
}

